//  Copyright (c) 2017-present The blackwidow Authors.  All rights reserved.
//  This source code is licensed under the BSD-style license found in the
//  LICENSE file in the root directory of this source tree. An additional grant
//  of patent rights can be found in the PATENTS file in the same directory.
//
#include <dirent.h>
#include <utility>

#include "blackwidow/backupable.h"

namespace blackwidow {

BackupEngine::~BackupEngine() {
  // Wait all children threads
  StopBackup();
  WaitBackupPthread();
  // Delete engines
  for (auto& engine : engines_) {
    delete engine.second;
  }
  engines_.clear();
}

Status BackupEngine::NewCheckpoint(rocksdb::DB* rocksdb_db,
                                   const std::string& type) {
  rocksdb::DBCheckpoint* checkpoint;
  Status s = rocksdb::DBCheckpoint::Create(rocksdb_db, &checkpoint);
  if (!s.ok()) {
    // log_warn("create checkpoint failed, error %s", s.ToString().c_str());
    return s;
  }
  engines_.insert(std::make_pair(type, checkpoint));
  return s;
}

Status BackupEngine::Open(blackwidow::BlackWidow *blackwidow,
                          BackupEngine** backup_engine_ptr) {
  *backup_engine_ptr = new BackupEngine();
  if (!*backup_engine_ptr) {
    return Status::Corruption("New BackupEngine failed!");
  }

  // Create BackupEngine for each db type
  rocksdb::Status s;
  rocksdb::DB *rocksdb_db;
  std::string types[] = {STRINGS_DB, HASHES_DB, LISTS_DB, ZSETS_DB, SETS_DB, EHASHES_DB};
  for (const auto& type : types) {
    if ((rocksdb_db = blackwidow->GetDBByType(type)) == NULL) {
      s = Status::Corruption("Error db type");
    }

    if (s.ok()) {
      s = (*backup_engine_ptr)->NewCheckpoint(rocksdb_db, type);
    }

    if (!s.ok()) {
      delete *backup_engine_ptr;
      break;
    }
  }
  return s;
}

Status BackupEngine::SetBackupContent() {
  Status s;
  for (const auto& engine : engines_) {
    // Get backup content
    BackupContent bcontent;
    s = engine.second->GetCheckpointFiles(bcontent.live_files,
        bcontent.live_wal_files,
        bcontent.manifest_file_size, bcontent.sequence_number);
    if (!s.ok()) {
      // log_warn("get backup files faild for type: %s", engine.first.c_str());
      return s;
    }
    backup_content_[engine.first] = std::move(bcontent);
  }
  return s;
}

Status BackupEngine::CreateNewBackupSpecify(const std::string &backup_dir,
                                            const std::string &type) {
  std::map<std::string, rocksdb::DBCheckpoint*>::iterator it_engine =
    engines_.find(type);
  std::map<std::string, BackupContent>::iterator it_content =
    backup_content_.find(type);
  std::string dir = GetSaveDirByType(backup_dir, type);
  delete_dir(dir.c_str());

  if (it_content != backup_content_.end() &&
      it_engine != engines_.end()) {
    Status s = it_engine->second->CreateCheckpointWithFiles(
        dir,
        it_content->second.live_files,
        it_content->second.live_wal_files,
        it_content->second.manifest_file_size,
        it_content->second.sequence_number);
    if (!s.ok()) {
      // log_warn("backup engine create new failed, type: %s, error %s",
      //    type.c_str(), s.ToString().c_str());
      return s;
    }

  } else {
    // log_warn("invalid db type: %s", type.c_str());
    return Status::Corruption("invalid db type");
  }
  return Status::OK();
}

void* ThreadFuncSaveSpecify(void *arg) {
  BackupSaveArgs* arg_ptr = static_cast<BackupSaveArgs*>(arg);
  BackupEngine* p = static_cast<BackupEngine*>(arg_ptr->p_engine);
  arg_ptr->res = p->CreateNewBackupSpecify(arg_ptr->backup_dir,
      arg_ptr->key_type);
  pthread_exit(&(arg_ptr->res));
}

Status BackupEngine::WaitBackupPthread() {
  int ret;
  Status s = Status::OK();
  for (auto& pthread : backup_pthread_ts_) {
    void *res;
    if ((ret = pthread_join(pthread.second, &res)) != 0) {
      // log_warn("pthread_join failed with backup thread for key_type:
      //    %s, error %d", pthread.first.c_str(), ret);
    }
    Status cur_s = *(static_cast<Status*>(res));
    if (!cur_s.ok()) {
      // log_warn("pthread executed failed with key_type: %s, error %s",
      //    pthread.first.c_str(), cur_s.ToString().c_str());
      StopBackup();  // stop others when someone failed
      s = cur_s;
    }
  }
  backup_pthread_ts_.clear();
  return s;
}


Status BackupEngine::CreateNewBackup(const std::string &dir) {
  Status s = Status::OK();
  std::vector<BackupSaveArgs*> args;
  for (const auto& engine : engines_) {
    pthread_t tid;
    BackupSaveArgs *arg = new BackupSaveArgs(
        reinterpret_cast<void*>(this), dir, engine.first);
    args.push_back(arg);
    if (pthread_create(&tid, NULL, &ThreadFuncSaveSpecify, arg) != 0) {
      s = Status::Corruption("pthead_create failed.");
      break;
    }
    if (!(backup_pthread_ts_.insert(
            std::make_pair(engine.first, tid)).second)) {
      // log_warn("thread open dupilicated, type: %s", engine.first.c_str());
      backup_pthread_ts_[engine.first] = tid;
    }
  }

  // Wait threads stop
  if (!s.ok()) {
    StopBackup();
  }
  s = WaitBackupPthread();

  for (auto& a : args) {
    delete a;
  }
  return s;
}

void BackupEngine::StopBackup() {
  // DEPRECATED
}

}  // namespace blackwidow
