//  Copyright (c) 2017-present The blackwidow Authors.  All rights reserved.
//  This source code is licensed under the BSD-style license found in the
//  LICENSE file in the root directory of this source tree. An additional grant
//  of patent rights can be found in the PATENTS file in the same directory.

#ifndef SRC_REDIS_STRINGS_H_
#define SRC_REDIS_STRINGS_H_

#include <string>
#include <vector>
#include <algorithm>

#include "rocksdb/utilities/titandb/db.h"

#include "src/redis.h"

namespace blackwidow {

class RedisStrings : public Redis {
 public:
  RedisStrings():Titandb_(nullptr){};
  ~RedisStrings(){ delete Titandb_; };

  // Common Commands
  Status Open(const BlackwidowOptions& bw_options,
              const std::string& db_path) override;
  Status ResetOption(const std::string& key, const std::string& value);
  Status CompactRange(const rocksdb::Slice* begin,
                      const rocksdb::Slice* end) override;
  Status GetProperty(const std::string& property, uint64_t* out) override;
  Status ScanKeyNum(uint64_t* num) override;
  Status ScanKeys(const std::string& pattern,
                  std::vector<std::string>* keys) override;

  // Strings Commands
  Status Append(const Slice& key, const Slice& value, int32_t* ret);
  Status BitCount(const Slice& key, int64_t start_offset, int64_t end_offset,
                  int32_t* ret, bool have_offset);
  Status BitOp(BitOpType op, const std::string& dest_key,
               const std::vector<std::string>& src_keys, int64_t* ret);
  Status Decrby(const Slice& key, int64_t value, int64_t* ret);
  Status Get(const Slice& key, std::string* value);
  Status GetWithTTL(const Slice& key, std::string* value, int64_t* timestamp);
  Status GetBit(const Slice& key, int64_t offset, int32_t* ret);
  Status Getrange(const Slice& key, int64_t start_offset, int64_t end_offset,
                  std::string* ret);
  Status GetrangeWithValue(const Slice& key, int64_t start_offset, int64_t end_offset,
                  std::string* ret, std::string* value, int64_t* ttl);
  Status GetSet(const Slice& key, const Slice& value, std::string* old_value);
  Status Incrby(const Slice& key, int64_t value, int64_t* ret);
  Status Incrbyfloat(const Slice& key, const Slice& value, std::string* ret);
  Status MGet(const std::vector<std::string>& keys,
              std::vector<ValueStatus>* vss);
  Status MSet(const std::vector<KeyValue>& kvs);
  Status MSetnx(const std::vector<KeyValue>& kvs, int32_t* ret);
  Status Set(const Slice& key, const Slice& value, const int32_t ttl = 0);
  Status Setxx(const Slice& key, const Slice& value,
               int32_t* ret, const int32_t ttl = 0);
  Status SetBit(const Slice& key, int64_t offset,
                int32_t value, int32_t* ret);
  Status Setex(const Slice& key, const Slice& value, int32_t ttl);
  Status Setnx(const Slice& key, const Slice& value,
               int32_t* ret, const int32_t ttl = 0);
  Status Setvx(const Slice& key, const Slice& value,
               const Slice& new_value, int32_t* ret, const int32_t ttl = 0);
  Status Delvx(const Slice& key, const Slice& value, int32_t* ret);
  Status Setrange(const Slice& key, int64_t start_offset,
                  const Slice& value, int32_t* ret);
  Status Strlen(const Slice& key, int32_t *len);

  Status BitPos(const Slice& key, int32_t bit, int64_t* ret);
  Status BitPos(const Slice& key, int32_t bit,
                int64_t start_offset, int64_t* ret);
  Status BitPos(const Slice& key, int32_t bit,
                int64_t start_offset, int64_t end_offset,
                int64_t* ret);
  Status PKScanRange(const Slice& key_start, const Slice& key_end,
                     const Slice& pattern, int32_t limit,
                     std::vector<KeyValue>* kvs, std::string* next_key);
  Status PKRScanRange(const Slice& key_start, const Slice& key_end,
                      const Slice& pattern, int32_t limit,
                      std::vector<KeyValue>* kvs, std::string* next_key);

  // Keys Commands
  Status Expire(const Slice& key, int32_t ttl) override;
  Status Del(const Slice& key) override;
  bool Scan(const std::string& start_key, const std::string& pattern,
            std::vector<std::string>* keys,
            int64_t* count, std::string* next_key) override;
  Status Expireat(const Slice& key, int32_t timestamp) override;
  Status Persist(const Slice& key) override;
  Status TTL(const Slice& key, int64_t* timestamp) override;

  // Iterate all data
  void ScanDatabase();

  rocksdb::titandb::TitanDB* GetTitandb() {
    return Titandb_;
  }
  
private:
  rocksdb::titandb::TitanDB *Titandb_;
};

}  //  namespace blackwidow
#endif  //  SRC_REDIS_STRINGS_H_
